/*
 * Copyright (c) 2003 Jens Mueller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
  	servlet class providing application's entry point by
  	implementing doGet and doPost

 	@author jens mueller
 */

import java.io.*;
import java.util.Vector;
import java.util.Iterator;
import java.util.StringTokenizer;

import javax.servlet.*;
import javax.servlet.http.*;

import org.dom4j.io.XMLWriter;
import org.dom4j.io.OutputFormat;

public class SnmpXmlGateway extends HttpServlet
{
	
private static String error = "";

/**
	sets up the <code>MibRepositoryHandler</code>
  	@see javax.servlet.GenericServlet#init()
 */
public void init()
{
	try
	{
		setUpConfig();
		
		// calling of getInstance() for setting up singleton class, all referenced schemas will be loaded
		MibRepositoryHandler handler = MibRepositoryHandler.getInstance();
	}
	catch ( Throwable th )
	{
		error = "init failed: "+th.toString();
	} 
}
	

/**
	reads the parameters from the depolyment descriptor, stores values
  	in <code>GatewayConf</code>
 
  	@throws GatewayConfigurationException
 */
private void setUpConfig() throws GatewayConfigurationException
{
	if ( !GatewayConf.isConfigured() )
	{
    	String repositoryURL;
    	String cachingTimeSecs;
    	String cachePurgeIntervalSecs;    	
    	long cachingSecs;
    	long cachingPurgeSecs;
    	String setAllowed;
		ServletContext context = getServletContext();
		
		try
		{
	    	repositoryURL				= context.getInitParameter( "repositoryURL" );
    		cachingTimeSecs 			= context.getInitParameter( "cachingTimeSecs" );
    		cachePurgeIntervalSecs	= context.getInitParameter( "cachePurgeIntervalSecs" );
    		setAllowed					= context.getInitParameter( "setAllowed" );	
		}
		catch ( Exception e )
		{
    		throw new GatewayConfigurationException( "Gateway misconfigured, " +
    																		"check for valid web.xml "+e.toString() );
		}

    	if ( repositoryURL == null || cachingTimeSecs == null || cachePurgeIntervalSecs == null )
    	{
    		throw new GatewayConfigurationException( "Gateway misconfigured, " +
    																		"check for existing and valid web.xml" );
    	}

    	try
    	{
    		cachingSecs 		= Long.parseLong( cachingTimeSecs );
    		cachingPurgeSecs	= Long.parseLong( cachePurgeIntervalSecs );
    		if ( cachingSecs < 1 || cachingPurgeSecs < 1 )
    		{
    			throw new GatewayConfigurationException("");
    		}
    	}
    	catch (Exception e)
    	{
    		throw new GatewayConfigurationException( "Gateway misconfigured, " +
    																		"check for valid web.xml" );
    	}
    	
    	GatewayConf.setRepositoryUrl( repositoryURL );
    	GatewayConf.setCachingTimeSecs( cachingSecs );    	
    	GatewayConf.setCachePurgerIntervalSecs( cachingPurgeSecs );
    	if( setAllowed != null && setAllowed.toUpperCase().equals( "YES" ) )
    		GatewayConf.setSetAllowed( true );
    	else
    		GatewayConf.setSetAllowed( false );    	
    	GatewayConf.setConfigured( true );
	}
}

	
/**
	handles get-requests submitted as value of HTTP-parameter "get".
  	steps are:
  	1. ensuring that deployment descriptor's params are loaded into <code>GatewayConf</code>
  	2. building a <code>GatewayGetRequest</code> via call of 
   	   <code>getGatewayGetRequest</code> of class <code>GatewayGetRequestFactory</code> 
  	3. obtaining the instance of the <code>MibRepositoryHandler</code>
  	4. executing request by calling <code>buildResponse</code> of at step 2 built
   	   <code>GatewayGetRequest</code>
  	5. writing resulting XML-document into servlet's response
  
  	This method is also main catchpoint of exceptions occured during processing. In occurance
  	of an exception, an accordant HTTP-Responsecode is sent, along with a describing
  	errormessage.
  	@see javax.servlet.http.HttpServlet#doGet(HttpServletRequest, HttpServletResponse)
 */
public void doGet( HttpServletRequest request, HttpServletResponse response )
    throws IOException, ServletException
    {
    	if ( error != "" )
		{
   			response.setContentType( "text/html" );
   			response.sendError( response.SC_INTERNAL_SERVER_ERROR, error );
    		return;
		}

		if ( !GatewayConf.isConfigured() )
		{
	    	try
    		{
    			setUpConfig();
    		}
	    	catch ( GatewayConfigurationException e )
    		{
    			response.setContentType( "text/html" );
    			response.sendError( response.SC_INTERNAL_SERVER_ERROR, e.getMessage() );
	    		return;
    		}
		}
    	
    	
    	if ( request.getParameter("get") == null || request.getParameter("get").length() == 0 )
    	{
    		/*
    		//debug
	   	    PrintWriter out = response.getWriter();
    		response.setContentType( "text/html" );
	   	   	out.println("<html><head><title>testpage for set</title></head>");
	   	   	out.println("<body>");
	   	   	out.println("<form name=\"setform\" method=\"post\" action=\"SERVLET-LOCATION\">");
	   	   	out.println("<textarea name=\"set\" rows=\"50\" cols=\"100\"></textarea>");
	   	   	out.println("<input type=\"submit\" value=\"go\"></form></body></html>");
    		//debug
	   	   	*/    		
    		response.setContentType( "text/html" );
    		response.sendError( response.SC_BAD_REQUEST, "no Request-String specified" );
    		return;
    	}
    	else
    	{
    		GatewayGetRequest gwrequest = null;
    		try 
    		{
    			gwrequest = GatewayGetRequestFactory.
    									getGatewayGetRequest( request.getParameter("get") );
    		}
    		catch ( MalformedRequestStringException e )
    		{
    			response.setContentType("text/html");
    			response.sendError( response.SC_BAD_REQUEST, e.getMessage() );
    			return;
    		}
    		
    		MibRepositoryHandler repHandler = null;
    		try
    		{
    			repHandler = MibRepositoryHandler.getInstance();
    		}
    		catch ( Throwable th )
    		{
    			response.setContentType( "text/html" );
    			response.sendError( 	response.SC_INTERNAL_SERVER_ERROR, 
												"Gateway misconfigured, init failed: " + th.getMessage() );
    			return;
    		}
  		 	try
    		 {
    		 	gwrequest.buildResponse( repHandler );
    		 }
    		 catch ( RequestExecutionException e )
    		 {
    			response.setContentType("text/html");
    			response.sendError( response.SC_NOT_FOUND, e.getMessage() );
    			return;
    		 }

   			response.setContentType("text/xml");    		 
	   	   	PrintWriter out = response.getWriter();
    		OutputFormat outputformat = new OutputFormat( "  ", true );
			XMLWriter xmlWriter = new XMLWriter( out, outputformat );
			xmlWriter.write( gwrequest.getResponseDocument() );
    		//debug
			//CacheHandler.dump( out );
    		//gwrequest.dump( out );
    		//repHandler.dump( out );
	   		//debug
    	}
    }
    
    
    /**
  	  	handles set-requests submitted as value of HTTP-parameter "set".
 	  	steps are:
 	  	1. ensuring that deployment descriptor's params are loaded into <code>GatewayConf</code>
 	  	2. checking if set-requests are allowed to perform
 	  	2. building a <code>GatewaySetRequest</code> via call of 
 	   	   <code>getGatewaySetRequest</code> of class <code>GatewaySetRequestFactory</code> 
 	  	3. executing request
  	  	5. sending an 202 - gateway processed the request HTTP-responsecode
 	  
 	  	This method is also main catchpoint of exceptions occured during processing. In occurance
 	  	of an exception, an accordant HTTP-Responsecode is sent, along with a describing
	  	errormessage.
 
      	@see javax.servlet.http.HttpServlet#doPost(HttpServletRequest, HttpServletResponse)
     */
    public void doPost( HttpServletRequest request, HttpServletResponse response )
    																	throws IOException, ServletException
	{
		String debug = "";
		
		if ( error != "" )
		{
   			response.setContentType( "text/html" );
   			response.sendError( response.SC_INTERNAL_SERVER_ERROR, error );
    		return;
		}

		if ( !GatewayConf.isConfigured() )
		{
	    	try
    		{
    			setUpConfig();
    		}
	    	catch ( GatewayConfigurationException e )
    		{
    			response.setContentType( "text/html" );
    			response.sendError( response.SC_INTERNAL_SERVER_ERROR, e.getMessage() );
	    		return;
    		}
		}
		
		if( !GatewayConf.isSetAllowed() )
		{
    			response.setContentType( "text/html" );
    			response.sendError( response.SC_FORBIDDEN , 	"service is only " + 
    																					"accepting get-requests" );			
    			return;
		}
		else
		{
			if( request.getParameter("set") == null )
	    	{
    			response.sendError( response.SC_BAD_REQUEST, 
																		"no valid XML-document transmitted" );
				return;
	    	}
    		else
    		{
	    		GatewaySetRequest gwRequest = null;
	    		//debug
		   	   	PrintWriter out = response.getWriter();
   				//response.setContentType("text/html");
		   	   	//out.println(request.getParameter("set"));
		   	   	//debug
    			try
    			{
	    			gwRequest = GatewaySetRequestFactory.getGatewaySetRequest( 
	    																					request.getParameter("set") );
    			}
    			catch ( MalformedSetRequestException e )
	    		{
    				response.setContentType("text/html");
    				response.sendError( response.SC_BAD_REQUEST, e.getMessage() );
    				return;	
	    		}
				
				try
				{
					gwRequest.execute();
				}
				catch( RequestExecutionException e )
				{
		    		//debug
		 	   		//response.setContentType("text/html");
		 	   		//out.println("<html><head></head><body>");	
					//gwRequest.dump( out );
		 	   		//out.println("</body></html>");	
		    		//debug
		 	   		response.setContentType("text/html");
    				response.sendError( response.SC_BAD_REQUEST, e.getMessage() );
    				return;
				}
   				response.sendError( response.SC_OK, "gateway processed the request" );
	    	}
    	}
	}
}